import java.awt.Dimension;
import java.io.PrintStream;
import java.nio.*;
import java.util.ArrayList;

import processing.core.*;
import processing.opengl.*;

import controlP5.ControlEvent;
import controlP5.ControlP5;

import peasy.*;
import peasy.org.apache.commons.math.geometry.Vector3D;

import toxi.geom.*;
import toxi.math.MathUtils;
import toxi.processing.*;

import javax.media.opengl.*;
import javax.media.opengl.glu.*;

/**
 * 
 * @author mannytan
 * @name NudiBranch
 * @date Dec 11, 2010
 * 
 */
public class NudiBranch extends PApplet {

	// Fan fan;
	Guides guides;
	PeasyCam cam;

	int totalSegments = 100;
	Segment firstSegment;
	ArrayList<Segment> segmentList = new ArrayList();

	private static PrintStream out = System.out;

	public static ToxiclibsSupport gfx;

	private Vec3D mouseVec;

	public static void main(String args[]) {
		PApplet.main(new String[] {  "NudiBranch" });
	}

	public void setup() {
		size(1200, 800, OPENGL);
		this.setSize(new Dimension(1200, 800)); // resize & set bounds of the
												// PApplet
		// this.setBounds(0, 0, 1200, 800);
		background(0);

		// Create Camera
		cam = new PeasyCam(this, 1000);
		cam.lookAt(0, 0, 0, 0);
		// Create Controls
		createControls();

		// Create ToxiclibsSupport
		gfx = new ToxiclibsSupport(this);

		// Create Guides
		guides = new Guides(this);

		// Create first segment data
		firstSegment = new Segment(this);
		firstSegment.seed = 0;
		firstSegment.seedNormal = 0.0f;
		firstSegment.parentSegment = firstSegment;
		// Create Segments
		for (int i = 0; i < totalSegments; i++) {
			Segment segment = new Segment(this);
			segment.seed = (i + 1);
			segment.seedNormal = (float) (i + 1) / (float) totalSegments;

			// if first segment, make parent itself
			if (i == 0) {
				segment.parentSegment = firstSegment;
			} else {
				segment.parentSegment = segmentList.get(segmentList.size() - 1);
			}

			segment.randomizeCurrentControls();
			segment.setPreviousControls();
			segmentList.add(segment);
		}

		gl = ((PGraphicsOpenGL) g).gl;
		glu = ((PGraphicsOpenGL) g).glu;

		mouseVec = new Ray3D();
	}

	public void draw() {
		// this.rotateY(frameCount*.01f);
		background(0);

		// camera HUD
		cam.beginHUD();
		guides.draw2DScope();
		controlP5.draw();
		updateControlParams();
		cam.endHUD();

		// draw guides
		if (isGuideVisible) {
			guides.draw3DAxes(400);
		}

		// create mouse pointer
		// *
		float sizer = 400.0f;
		float speeder = frameCount * .0025f;
		mouseVec.x = sin(MathUtils.PI * speeder * .5f) * sizer + 0.0f;
		mouseVec.y = cos(MathUtils.PI * speeder) * sizer + 0.0f;
		mouseVec.z = sin(MathUtils.PI * speeder) * sizer + 0.0f;
		// */
		/*
		 * float[] mousePosList = getMouse3D();
		 * 
		 * mouseVec.x += (mousePosList[0] - mouseVec.x) * .05f; mouseVec.y +=
		 * (mousePosList[1] - mouseVec.y) * .05f; mouseVec.z += (mousePosList[2]
		 * - mouseVec.z) * .05f;
		 * 
		 * mouseVec.normalizeTo(300.0f);
		 * 
		 * //
		 */

		// limiters
		/*
		 * if (mouseVec.z < -2000){ mouseVec.z = -2000; } if (mouseVec.z >
		 * 2000){ mouseVec.z = 2000; } //
		 */

		setLineStyle(5, 0xFFFF6600, 1.0f);
		gfx.point(mouseVec);

		// first segment
		firstSegment.current.set(mouseVec);
		firstSegment.previous.set(mouseVec);
		firstSegment.parse();

		for (Segment segment : segmentList) {
			segment.parse();
			segment.draw(gfx);
		}

	}

	public void keyPressed() {
		if (key == ' ') {
			reset();
		}
	}

	public void reset() {

		for (Segment segment : segmentList) {
			segment.randomizeCurrentControls();
			segment.setPreviousControls();
		}

	}

	GL gl;
	GLU glu;

	float[] getMouse3D() {
		((PGraphicsOpenGL) g).beginGL();
		int viewport[] = new int[4];
		double[] proj = new double[16];
		double[] model = new double[16];
		gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
		gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, proj, 0);
		gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, model, 0);
		FloatBuffer fb = ByteBuffer.allocateDirect(4).order(ByteOrder.nativeOrder()).asFloatBuffer();
		gl.glReadPixels(mouseX, height - mouseY, 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT, fb);
		fb.rewind();
		double[] mousePosArr = new double[4];
		glu.gluUnProject((double) mouseX, height - (double) mouseY, (double) fb.get(0), model, 0, proj, 0, viewport, 0, mousePosArr, 0);
		((PGraphicsOpenGL) g).endGL();
		return new float[] { (float) mousePosArr[0], (float) mousePosArr[1], (float) mousePosArr[2] };
	}

	/**
	 ***************************************************************************** 
	 * @TODO Parent child relation is screwy. controls are in main class but
	 *       will be migrated to it's own object eventually
	 ***************************************************************************** 
	 */
	// control params
	ControlP5 controlP5;
	private boolean isGuideVisible;
	public float rootScalar = 10.0f;
	public float rootFrameCountScalar = 0.0f;
	public float rootSeedNormalScalar = 0.0f;

	public float rootLayerRadiusOffset = 0.0f;
	public float rootLayerRadiusScalar = 250.0f;
	public float rootLayerSpeedOffset = 5.0f;

	public float rootFlappingSpeedOffset = -0.04f;
	public float rootFlappingSpeedNormal = 0.02f;
	public float rootMaxAngle = 180.0f;

	private void updateControlParams() {

		for (Segment segment : segmentList) {
			segment.segmentScalar = rootScalar;
			segment.frameCount = frameCount;
			segment.frameCountScalar = rootFrameCountScalar;
			segment.seedNormalScalar = rootSeedNormalScalar;

			segment.layerRadiusOffset = rootLayerRadiusOffset;
			segment.layerRadiusScalar = rootLayerRadiusScalar;
			segment.layerSpeedOffset = rootLayerSpeedOffset;

			segment.flappingSpeedNormal = rootFlappingSpeedNormal;
			segment.flappingSpeedOffset = rootFlappingSpeedOffset;
			segment.maxAngle = rootMaxAngle;

		}
	}

	private void createControls() {
		isGuideVisible = true;

		controlP5 = new ControlP5(this);
		controlP5.setAutoDraw(false);

		int offsetX = 10;
		int offsetY = 10;

		int buttonWidth = 100;
		int buttonHeight = 20;
		int buttonMargin = 5;

		int sliderWidth = 190;
		int sliderHeight = 12;
		int sliderHeightSpread = 20;

		controlP5.addToggle("toggleIsGuideVisible", isGuideVisible, offsetX, offsetY, buttonWidth, buttonHeight);
		offsetY += (buttonHeight + buttonMargin + 20);
		controlP5.addSlider("rootScalar", 0, 100, rootScalar, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		controlP5.addSlider("rootFrameCountScalar", -.25f, .25f, rootFrameCountScalar, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		controlP5.addSlider("rootSeedNormalScalar", 0.0f, 360.0f, rootSeedNormalScalar, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		offsetY += 20;
		controlP5.addSlider("rootLayerRadiusOffset", 0.0f, 500.0f, rootLayerRadiusOffset, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		controlP5.addSlider("rootLayerRadiusScalar", -500.0f, 500.0f, rootLayerRadiusScalar, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		controlP5.addSlider("rootLayerSpeedOffset", 0.0f, 15.0f, rootLayerSpeedOffset, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);

		offsetY += 20;
		controlP5.addSlider("rootFlappingSpeedNormal", 0.0f, 1.0f, rootFlappingSpeedNormal, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		controlP5.addSlider("rootFlappingSpeedOffset", -1.0f, 1.0f, rootFlappingSpeedOffset, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		offsetY += 20;

		controlP5.addSlider("rootMaxAngle", 0.0f, 360.0f, rootMaxAngle, offsetX, offsetY, sliderWidth, sliderHeight);
		offsetY += (buttonHeight + buttonMargin);
		offsetY += 20;

	}

	public void toggleIsGuideVisible(boolean theFlag) {
		isGuideVisible = theFlag;
	}

	public void controlEvent(ControlEvent theEvent) {
		cam.setActive(false);
	}

	public void mouseReleased() {
		cam.setActive(true);
	}

	public void setLineStyle(int weightParam, int colorParam, float alphaParam) {
		this.noFill();
		this.strokeWeight(weightParam);
		this.stroke(colorParam, 255.0f * alphaParam);

	}
}
