import java.util.ArrayList;

import processing.core.PApplet;
import toxi.geom.*;
import toxi.math.MathUtils;
import toxi.processing.ToxiclibsSupport;

public class Segment extends BoilerPlate {
	PApplet parent;

	public int seed = 0;
	public float seedNormal = 0.0f;

	public Ray3D currentRotator;
	public Ray3D currentArbitrary;
	public Ray3D currentTangent;

	public Vec3D currentDirection;
	public Ray3D current;

	public Vec3D previousDirection;
	public Ray3D previous;

	public Segment parentSegment;

	// static params
	public int totalWings = 2;
	public int totalLayers = 20;

	// dynamic params
	public int frameCount;
	public float frameCountScalar;
	public float segmentScalar;
	public float seedNormalScalar;
	public float layerRadiusOffset;
	public float layerRadiusScalar;
	public float layerSpeedOffset;
	public float flappingSpeedOffset;
	public float flappingSpeedNormal;
	public float maxAngle;


	public Vec3D[][] segmentPoints = new Vec3D[totalWings][totalLayers];

	Segment(PApplet p) {
		parent = p;
		setScope("Segment");
		currentDirection = new Vec3D();
		current = new Ray3D();
		previousDirection = new Vec3D();
		previous = new Ray3D();

		currentArbitrary = new Ray3D(new Vec3D(), Vec3D.Y_AXIS);
		currentTangent = new Ray3D();
		currentRotator = new Ray3D();

		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				segmentPoints[i][j] = new Vec3D();
			}
		}

		// tracer(segmentPoints[0].get(0));
	}

	public void randomizeCurrentControls() {
		float r;
		r = 1.0f;
		currentDirection = new Vec3D();
		r = 100.0f;
		current = new Ray3D();
	}

	public void setPreviousControls() {
		previousDirection = parentSegment.currentDirection;
		previous = parentSegment.current;
	}

	public void parse() {
		previousDirection.set(parentSegment.currentDirection);
		previous.set(parentSegment.current);

		current.set(current.sub(previous));
		current.normalizeTo(segmentScalar);
		current.addSelf(previous);
		current.setDirection(previous.sub(current));

		currentArbitrary.set(current);
		currentTangent.set(current);
		currentRotator.set(current);

		Vec3D currentVec = new Vec3D(current);
		Vec3D dirVec = new Vec3D(current.getDirection());
		Vec3D crossVec = currentArbitrary.cross(dirVec.getNormalized()).normalize();

		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				// convert positions
				float wingPercentage = (float) i / (float) totalWings;
				float layerPercentage = (float) j / (float) totalLayers;

				float angleInRads = MathUtils.TWO_PI * wingPercentage;
				angleInRads += seedNormal * seedNormalScalar; // extra offset of
																// rotation
				angleInRads += frameCount * frameCountScalar; // speed of
																// rotation

				float flappingSpeed = flappingSpeedNormal;
				flappingSpeed += seedNormal * flappingSpeedOffset * .01f;
				flappingSpeed *= frameCount;
				flappingSpeed += (1 - layerPercentage) * layerSpeedOffset;

				// flips it around based on 0,1 --> -1,1
				float flappingRange = maxAngle * (float) ((i % 2) * 2 - 1);

				float rangeOffsetInRads = MathUtils.sin(flappingSpeed) * flappingRange / 360.0f * MathUtils.PI;

				angleInRads += (rangeOffsetInRads * (layerPercentage));

				// create local rotation vector
				Vec3D rotVec = crossVec.getRotatedAroundAxis(dirVec, angleInRads);

				// extra offset of rotation
				float layerRadius = layerRadiusScalar * layerPercentage;
				layerRadius += layerRadiusOffset;
				layerRadius *= MathUtils.sin(MathUtils.PI * seedNormal);

				Vec3D point = new Vec3D(rotVec);
				point.scaleSelf(layerRadius);
				point.addSelf(currentVec);

				segmentPoints[i][j].set(point);
			}
		}
	}

	public void draw(ToxiclibsSupport gfx) {
		// tracer("draw");
		parent.noFill();
		float wingPercentage;
		float layerPercentage;

		/*
		// orange line down the center
		float colorAlpha = (1 - seedNormal) * 255.0f + 0.0f;
		setLineStyle(1, 0xFFFF6600, (1 - seedNormal));
		gfx.ray(current, segmentScalar);
		//*/
		/*
		// yellow dots
		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;

				Vec3D point = new Vec3D(segmentPoints[i][j]);
				float lineAlpha = (1 - seedNormal * 0) * (1 - layerPercentage);

				setLineStyle(1, 0xFFFFFF00, lineAlpha);
				gfx.point(point);
			}
		}
		//*/
		for (int i = 0; i < totalWings; i++) {
			for (int j = 0; j < totalLayers; j++) {
				wingPercentage = (float) i / (float) totalWings;
				layerPercentage = (float) j / (float) totalLayers;

				float lineAlpha = (1 - seedNormal * 0) * (1 - layerPercentage);

				Vec3D a = new Vec3D(segmentPoints[i][j]);
				a.add(current);
				Vec3D b = new Vec3D(parentSegment.segmentPoints[i][j]);
				b.add(previous);

				setLineStyle(1, 0xFF00FFFF, lineAlpha);
				gfx.line(a,b);
			}
		}

	}

	public void setLineStyle(int weightParam, int colorParam, float alphaParam) {
		parent.noFill();
		parent.strokeWeight(weightParam);
		parent.stroke(colorParam, 255.0f * alphaParam);

	}
}
